package perf

import (
	"io"
	"os"
	"path/filepath"
	"runtime"
	"testing"

	"github.com/jarcoal/httpmock"
	"github.com/layer5io/meshery/mesheryctl/pkg/utils"
)

var tempProfileID = "a2a555cf-ae16-479c-b5d2-a35656ba741e"

// golden file responses
var (
	// standard api response for abhishek profile
	result1000 = "1000.golden"
	// standard api response of 25 performance result
	result1001 = "1001.golden"
	// api response of zero performance result
	result1004 = "1004.golden"
	// unable marshal response
	result1005 = "1005.golden"
	// empty response
	result1006 = "1006.golden"
)

// golden file mesheryctl outputs
var (
	// mesheryctl response of 25 performance result
	result1001output = "1001.golden"
	// mesheryctl response of 25 performance result in expanded output
	result1004output = "1004.golden"
	// mesheryctl response when no results found
	result1005output = "1005.golden"
	// mesheryctl response of 25 performance result in json output
	result1006output = "1006.golden"
	// mesheryctl response of 25 performance result in yaml output
	result1007output = "1007.golden"
	// mesheryctl response for invalid output format
	result1008output = "1008.golden"
	// mesheryctl response for server response code error
	result1009output = "1009.golden"
	// mesheryctl response for unmarshal error
	result1010output = "1010.golden"
	// mesheryctl response for failing attach authentication
	result1011output = "1011.golden"
	// mesheryctl response for no profile-id passed
	result1012output = "1012.golden"
)

func TestResultCmd(t *testing.T) {
	utils.SetupContextEnv(t)
	utils.StartMockery(t)
	testContext := utils.NewTestHelper(t)

	// get current directory
	_, filename, _, ok := runtime.Caller(0)
	if !ok {
		t.Fatal("Not able to get current working directory")
	}
	currDir := filepath.Dir(filename)
	fixturesDir := filepath.Join(currDir, "fixtures", "result")
	testToken := filepath.Join(currDir, "fixtures", "auth.json")
	testdataDir := filepath.Join(currDir, "testdata", "result")

	profileURL := testContext.BaseURL + "/api/user/performance/profiles"
	resultURL := testContext.BaseURL + "/api/user/performance/profiles/" + tempProfileID + "/results"

	tests := []tempTestStruct{
		{"standard results output", []string{"result", "abhishek"}, []utils.MockURL{
			{Method: "GET", URL: profileURL, Response: result1000, ResponseCode: 200},
			{Method: "GET", URL: resultURL, Response: result1001, ResponseCode: 200},
		}, result1001output, testToken, false},
		{"Unmarshal error", []string{"result", "abhishek"}, []utils.MockURL{
			{Method: "GET", URL: profileURL, Response: result1000, ResponseCode: 200},
			{Method: "GET", URL: resultURL, Response: result1005, ResponseCode: 200},
		}, result1010output, testToken, true},
		{"failing add authentication test", []string{"result", "abhishek"}, []utils.MockURL{}, result1011output, testToken + "invalid-path", true},
		{"Server Error 500", []string{"result", "abhishek"}, []utils.MockURL{
			{Method: "GET", URL: profileURL, Response: result1000, ResponseCode: 200},
			{Method: "GET", URL: resultURL, Response: result1006, ResponseCode: 500},
		}, result1009output, testToken, true},
		{"No profile passed", []string{"result"}, []utils.MockURL{}, result1012output, testToken, true},
	}

	testsforLogrusOutputs := []tempTestStruct{
		{"standard results in json output", []string{"result", "abhishek", "-o", "json"}, []utils.MockURL{
			{Method: "GET", URL: profileURL, Response: result1000, ResponseCode: 200},
			{Method: "GET", URL: resultURL, Response: result1001, ResponseCode: 200},
		}, result1006output, testToken, false},
		{"standard results in yaml output", []string{"result", "abhishek", "-o", "yaml"}, []utils.MockURL{
			{Method: "GET", URL: profileURL, Response: result1000, ResponseCode: 200},
			{Method: "GET", URL: resultURL, Response: result1001, ResponseCode: 200},
		}, result1007output, testToken, false},
		{"invalid output format", []string{"result", "abhishek", "-o", "invalid"}, []utils.MockURL{
			{Method: "GET", URL: profileURL, Response: result1000, ResponseCode: 200},
			{Method: "GET", URL: resultURL, Response: result1001, ResponseCode: 200},
		}, result1008output, testToken, true},
	}

	// Run tests in list format
	for _, tt := range tests {
		t.Run(tt.Name, func(t *testing.T) {
			utils.TokenFlag = tt.Token

			for _, mock := range tt.URLs {
				apiResponse := utils.NewGoldenFile(t, mock.Response, fixturesDir).Load()
				httpmock.RegisterResponder(mock.Method, mock.URL,
					httpmock.NewStringResponder(mock.ResponseCode, apiResponse))
			}

			golden := utils.NewGoldenFile(t, tt.ExpectedResponse, testdataDir)
			_ = utils.SetupMeshkitLoggerTesting(t, false)

			// Grab console prints
			rescueStdout := os.Stdout
			r, w, _ := os.Pipe()
			os.Stdout = w

			PerfCmd.SetArgs(tt.Args)
			PerfCmd.SetOutput(rescueStdout)
			err := PerfCmd.Execute()
			if err != nil {
				if tt.ExpectError {
					if *update {
						golden.Write(err.Error())
					}
					expectedResponse := golden.Load()
					utils.Equals(t, expectedResponse, err.Error())
					resetVariables()
					return
				}
				t.Error(err)
			}

			w.Close()
			out, _ := io.ReadAll(r)
			os.Stdout = rescueStdout

			// response being printed in console
			actualResponse := string(out)
			// write it in file
			if *update {
				golden.Write(actualResponse)
			}
			expectedResponse := golden.Load()
			utils.Equals(t, expectedResponse, actualResponse)
			resetVariables()
		})
	}

	// Run tests in list format
	for _, tt := range testsforLogrusOutputs {
		t.Run(tt.Name, func(t *testing.T) {
			utils.TokenFlag = tt.Token

			for _, mock := range tt.URLs {
				apiResponse := utils.NewGoldenFile(t, mock.Response, fixturesDir).Load()
				httpmock.RegisterResponder(mock.Method, mock.URL,
					httpmock.NewStringResponder(mock.ResponseCode, apiResponse))
			}

			golden := utils.NewGoldenFile(t, tt.ExpectedResponse, testdataDir)

			b := utils.SetupMeshkitLoggerTesting(t, false)

			PerfCmd.SetArgs(tt.Args)
			PerfCmd.SetOutput(b)
			err := PerfCmd.Execute()
			if err != nil {
				if tt.ExpectError {
					if *update {
						golden.Write(err.Error())
					}
					expectedResponse := golden.Load()
					utils.Equals(t, expectedResponse, err.Error())
					resetVariables()
					return
				}
				t.Error(err)
			}

			// response being printed in console
			actualResponse := b.String()
			// write it in file
			if *update {
				golden.Write(actualResponse)
			}
			expectedResponse := golden.Load()
			utils.Equals(t, expectedResponse, actualResponse)
			resetVariables()
		})
	}

	// stop mock server
	utils.StopMockery(t)
}
